import numpy as np
from pypka.config import Config
from pypka.concurrency import runMCCalcs, startPoolProcesses
from pypka.constants import KBOLTZ, PKAPLACEHOLDER, MAXNPKHALFS


def calcpKhalfs(pH, nsites, avgs, pmean, pKs, dpH):
    for site in range(nsites):
        mean = avgs[site]
        p = pmean[site]

        if p > 0.5 and mean <= 0.5 or p < 0.5 and mean >= 0.5:
            pKhalf = pH - dpH * (mean - 0.5) / (mean - p)
            for i in range(MAXNPKHALFS):
                if pKs[site][i] == PKAPLACEHOLDER:
                    pKs[site][i] = pKhalf
                    break

        elif p > 1.5 and mean <= 1.5 or p < 1.5 and mean >= 1.5:
            pKhalf = pH - dpH * (mean - 1.5) / (mean - p)
            for i in range(MAXNPKHALFS):
                if pKs[site][i] == PKAPLACEHOLDER:
                    pKs[site][i] = pKhalf
                    break
    return pKs


class MonteCarlo:
    """Monte Carlo Simulation"""

    def __init__(self, sites):
        self.sites = sites
        self.nsites = len(sites)

        self.init_mc_vars()
        self.run()
        self.get_pkas()
        self.get_tit_states()
        self.get_coupled_sites()

    def init_mc_vars(self):
        """Initialize need variables"""

        def resize_list_of_lists(listn, maxsize, filler=None):
            for i in listn:
                diff = maxsize - len(i)
                for _ in range(diff):
                    i.append(filler)

        possible_states_g = [[] for _ in self.sites]
        possible_states_occ = [[] for _ in self.sites]

        temperature = float(Config.pypka_params["temp"])
        isite = -1
        for site in self.sites:
            isite += 1
            itaut = 0
            for tautomer in site.iterOrderedTautomersWithoutRef():
                dg = tautomer.dg / (KBOLTZ * temperature)
                possible_states_g[isite].append(dg)

                if site.type == "c":
                    prot_state = 0
                elif site.type == "a":
                    prot_state = 1

                possible_states_occ[isite].append(prot_state)
                itaut += 1

            if site.type == "c":
                prot_state = 1
            elif site.type == "a":
                prot_state = 0
            possible_states_occ[isite].append(prot_state)
            possible_states_g[isite].append(0.0)

        maxstates = max(Config.parallel_params.npossible_states)
        interactions_look = Config.parallel_params.interactions_look

        resize_list_of_lists(possible_states_g, maxstates)
        resize_list_of_lists(possible_states_occ, maxstates, filler=-500)
        resize_list_of_lists(interactions_look, maxstates, filler=-500)

        Config.parallel_params.possible_states_g = possible_states_g
        Config.parallel_params.possible_states_occ = possible_states_occ

    def run(self):
        """Run a MC sim"""
        params = Config.mc_params
        self.dpH = params["pHstep"]
        self.pH_values = params.pH_values

        ncpus = min(Config.pypka_params["ncpus"], self.nsites)
        results = startPoolProcesses(
            runMCCalcs,
            self.pH_values,
            ncpus,
            assign="ordered",
            merged_results=True,
        )

        self.counts_all = []
        self.avgs_all = []
        self.final_states = []
        self.coupled_sites_pairs = []

        for i in results:
            self.avgs_all.append(i[0])
            self.counts_all.append(i[1])
            self.final_states.append(i[3])
            self.coupled_sites_pairs.append(i[4])
        self.coupled_sites_pairs = self.coupled_sites_pairs[0]

    def get_pkas(self):
        mcsteps = Config.mc_params["mcsteps"]

        pKs = np.array(
            [[PKAPLACEHOLDER for ii in range(MAXNPKHALFS)] for i in range(self.nsites)]
        )

        self.total_tit_curve = {}
        for pHstep, pH in enumerate(self.pH_values):
            ph_avgs_prots = self.avgs_all[pHstep] / mcsteps
            self.total_tit_curve[pH] = sum(ph_avgs_prots) / self.nsites

            if pHstep > 0:
                pKs = calcpKhalfs(
                    pH, self.nsites, ph_avgs_prots, prev_prots, pKs, self.dpH
                )

            prev_prots = ph_avgs_prots[:]

        pKas = pKs

        text_pks = ""
        text_prots = "#pH      "
        c = -1
        for i in pKas:
            c += 1
            pK = i[0]
            site = self.sites[c]
            site.setpK(pK)
            chain = site.molecule.chain
            sitename = site.getName()
            resnumb = site.getResNumber(correct_icode=True)
            text_prots += "{0}_{2}_{1} ".format(resnumb, sitename, chain)
            text_pks += "{0:>5} {1:3} {2:20} {3:3}\n".format(
                resnumb, sitename, str(pK), chain
            )
        self.text_pks = text_pks
        self.text_prots = text_prots

    def get_tit_states(self):
        mcsteps = Config.mc_params["mcsteps"]

        for pHstep, pH in enumerate(self.pH_values):
            self.text_prots += "\n{pH:5.2f}".format(pH=pH)

            ph_avgs_prots = self.avgs_all[pHstep] / mcsteps
            ph_taut_probs = self.counts_all[pHstep] / mcsteps

            for c, site in enumerate(self.sites):
                mean = ph_avgs_prots[c]
                res_state_dist = list(ph_taut_probs[c])

                ntauts = site.getNTautomers()
                ref_i = ntauts
                prot_state = site.getRefProtState()
                if (mean > 0.5 and prot_state == 1) or (
                    mean <= 0.5 and prot_state == -1
                ):
                    state_i = ref_i
                else:
                    max_prob = max(res_state_dist[:ref_i])
                    state_i = res_state_dist.index(max_prob)

                most_prob_state = state_i + 1

                if mean != PKAPLACEHOLDER:
                    self.text_prots += "\t{mean:7.4f}".format(mean=mean)
                else:
                    self.text_prots += "\t-"

                site.most_prob_states[pH] = most_prob_state
                site.final_states[pH] = self.final_states[pHstep][c]
                site.tit_curve[pH] = mean
                site.states_prob[pH] = res_state_dist

    def get_coupled_sites(self):
        """Get coupled sites"""
        self.coupled_sites_dict = {}
        for sitei1, sitei2 in self.coupled_sites_pairs:
            site1 = Config.parallel_params.all_sites[sitei1]
            site2 = Config.parallel_params.all_sites[sitei2]

            if site1 not in self.coupled_sites_dict:
                self.coupled_sites_dict[site1] = []
            if site2 not in self.coupled_sites_dict:
                self.coupled_sites_dict[site2] = []
            self.coupled_sites_dict[site1].append(site2)
            self.coupled_sites_dict[site2].append(site1)

            to_print = "{:3} {:3} {:6} | {:3} {:3} {:6}".format(
                site1.molecule.chain,
                site1.res_name,
                site1.res_number,
                site2.molecule.chain,
                site2.res_name,
                site2.res_number,
            )
            print(to_print)

        self.text_coupled_sites = ""
        for site1, sites in self.coupled_sites_dict.items():
            to_print = "{:3} {:3} {:6} | {:6} -> ".format(
                site1.molecule.chain, site1.res_name, site1.res_number, len(sites)
            )

            for site2 in sites:
                to_print += "{:3} {:3} {:6}, ".format(
                    site2.molecule.chain, site2.res_name, site2.res_number
                )
            self.text_coupled_sites += to_print[:-2] + "\n"



import argparse
import os

from pypka.main import Titration
from pypka.__init__ import __version__


def read_settings(filename):
    """
    Reads the settings file.

    This file should have the following format:
      - commented lines should being with a '#'
      - every parameter should be declared as such: name = value

    All parameter values are interpreted as strings, however, a type
        check is later performed for each declared input value.
    All parameter names not recognizable are reported as a warning.
    """
    parameters = {}
    parameters["lipid_definition"] = {}
    with open(filename) as f:
        nline = 0
        for line in f:
            nline += 1
            line = line.strip()
            if line and line[0] != "#":
                no_comments = line.split("#")[0]
                parts = no_comments.split("=")
                param_name = parts[0].strip()
                param_value = "=".join(parts[1:]).strip()
                if "lipid_definition" in param_name:
                    parts = param_value.split(":")
                    old_name = parts[0]
                    new_name = parts[1]
                    parameters["lipid_definition"][old_name] = new_name
                elif len(parts) != 2 or not param_name or not param_value:
                    raise IOError(
                        "Incorrect format in line {0} of file {1}: "
                        "\n{1}#{0}: {2}".format(nline, filename, line)
                    )
                else:
                    parameters[param_name] = param_value
    # Search for all titrable sites in different chains
    sites = {}
    for param_name in parameters:
        if param_name.startswith("site"):
            if "_" in param_name:
                chain = param_name.split("_")[1]
            else:
                chain = " "
            sites_str = parameters[param_name]
            chain_sites = [i.strip() for i in sites_str.split(",")]
            sites[chain] = chain_sites

    if " " in sites and sites[" "] == ["all"]:
        if len(sites) != 1:
            raise Exception(
                '"sites" parameter is incorrectly defined.\n '
                'Incompatible parameters: "sites" and the multi-chain nomenclature "sites_X" where X is the chain'
                "Please choose one of the two ways to define titratble sites.\n"
                "Automatic Example: sites = all\n"
                "Residues NTR, 1, 5 and 8 in chain A \nExample: sites_A = 1N, 1, 5, 8"
            )
        sites = "all"

    if not sites:
        raise Exception(
            'Missing mandatory "sites" parameter\nTitratate all sites: sites = all\nTitrate all sites only from chain A: sites_A = all'
        )

    return sites, parameters


def check_cli_args():
    """Gets the CLI arguments and interprets them"""
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""
PypKa
A python module for flexible Poisson-Boltzmann based pKa calculations with proton tautomerism

Documentation can be found at https://pypka.readthedocs.io/en/latest/

PypKa is distributed under a LGPL-3.0, however delphi4py depends on DelPhi which is proprietary.
To use DelPhi the user is required to download the DelPhi license https://honiglab.c2b2.columbia.edu/software/cgi-bin/software.pl?input=DelPhi

Example:
python3 pypka.py params.dat --debug""",
    )

    # Mandatory Arguments
    parser.add_argument(
        "settings", help=" settings file name", default="settings.dat", action="store"
    )

    # Optional Arguments
    parser.add_argument(
        "--debug",
        help="activation of the debug mode " "to print extra information",
        action="store_true",
    )

    parser.add_argument("--version", action="version", version=__version__)

    args = parser.parse_args()

    # Apply some criteria to input arguments
    if not os.path.isfile(args.settings):
        raise IOError("File {0} does not exist.".format(args.settings))

    # Read Settings File
    sites, parameters = read_settings(args.settings)

    return sites, parameters, args.debug


def CLI():
    # Read command line arguments
    sites, parameters, debug = check_cli_args()

    Titration(parameters, sites=sites, debug=debug)
    print("CLI exited successfully")



from copy import copy
from pypka.config import Config
from pypka.constants import (
    TERMINAL_OFFSET,
    PROTEIN_RESIDUES,
    TITRABLERESIDUES,
    TITRABLETAUTOMERS,
    REGULARTITRATINGRES,
)
from pypka.clean.ffconverter import main_chains
from pdbmender.formats import new_pdb_line, read_pdb_line
import logging

logger = logging.getLogger(__name__)


def warn_integrity_check_failed(resnumb, resname):
    warn = "{0} {1} failed integrity check".format(resnumb, resname)
    logger.warning(warn)


def addSiteAndTautomers(molecule, resnumb, resname, termini_resname=None):
    if resname in TITRABLETAUTOMERS:
        ntautomers = TITRABLETAUTOMERS[resname]
    else:
        for res in REGULARTITRATINGRES:
            if res[0:2] == resname[0:2]:
                ntautomers = TITRABLETAUTOMERS[res]
    sID = molecule.addSite(resnumb)
    molecule.addTautomers(sID, ntautomers, resname, termini_resname=termini_resname)


def check_CYS_bridge(res_atoms, resnumb):
    if Config.pypka_params["ff_family"] == "GROMOS":
        CYS_atoms = ["N", "CA", "CB", "SG", "C", "O", "H"]
    elif Config.pypka_params["ff_family"] == "CHARMM":
        CYS_atoms = ["N", "CA", "CB", "SG", "C", "O", "HN", "HA", "HB1", "HB2"]

    if set(res_atoms).issubset(CYS_atoms) and set(CYS_atoms).issubset(res_atoms):
        warn = "CYS-{0} is assumed to be participating in a SS-bond".format(resnumb)
        logger.warning(warn)
        return True
    else:
        return False


def check_site(molecule, resnumb, sites, resname, cur_atoms, ter=None):
    # Corrects weird tautomer residue names like GL4
    if resname not in PROTEIN_RESIDUES:
        for res in REGULARTITRATINGRES:
            if res[0:2] == resname[0:2]:
                resname = res

    is_titrating = True
    if ter:
        if not Config.pypka_params["ser_thr_titration"] and resname in (
            "SER",
            "THR",
        ):
            is_titrating = False
        else:
            is_titrating = bool(resname in TITRABLERESIDUES)

    elif not Config.pypka_params["ser_thr_titration"] and resname in (
        "SER",
        "THR",
    ):
        return

    res_atoms = copy(cur_atoms)
    (integrity_terminal, integrity_site) = check_integrity(
        resname, res_atoms, is_titrating, ter=ter
    )

    if integrity_terminal:
        ter_resnumb = resnumb + TERMINAL_OFFSET
        addSiteAndTautomers(molecule, ter_resnumb, ter, termini_resname=resname)
        if ter == "NTR":
            molecule.NTR += [resnumb]
        elif ter == "CTR":
            molecule.CTR += [resnumb]
    elif ter:
        warn_integrity_check_failed(resnumb, ter)

    # print(resname, resnumb, integrity_site, resnumb in molecule.CYS_bridges)

    if resnumb in sites:
        if integrity_site:
            addSiteAndTautomers(molecule, resnumb, resname)
            if resname == "CYS":
                molecule.correct_names[resnumb] = "CY0"
        else:
            warn_integrity_check_failed(resnumb, resname)
    elif resnumb in molecule.CYS_bridges or resname == "CYS":
        if ter == "NTR":
            cur_atoms = set(cur_atoms) - set(["H1", "H2", "H3"])
            cur_atoms.add("H")
        elif ter == "CTR":
            cur_atoms = set(cur_atoms) - set(
                ["HO11", "HO12", "HO21", "HO22", "O1", "O2"]
            )
            cur_atoms.update(("H", "O"))

        is_cys_bridge = check_CYS_bridge(cur_atoms, resnumb)
        if not is_cys_bridge and resnumb in molecule.CYS_bridges:
            warn_integrity_check_failed(resnumb, resname)
        elif is_cys_bridge and Config.pypka_params["ff_family"] == "CHARMM":
            molecule.correct_names[resnumb] = "SSB"


def check_sites_integrity(filename, molecules, chains_res):
    """Identifies titrable residues and checks integrity of the residue blocks
    (excluding Hydrogens)
    """
    resnumb = None
    cur_atoms = []
    prev_resnumb = None
    prev_resname = None
    last_chain = None
    chain = None
    with open(filename) as f:
        nline = 0
        f_lines = f.readlines()
        maxnlines = len(f_lines)
        for line in f_lines:
            resname = None
            nline += 1

            if "ATOM " == line[0:5]:
                (aname, anumb, resname, chain, resnumb, x, y, z) = read_pdb_line(line)

                if last_chain in molecules or (chain in molecules and not last_chain):
                    if not last_chain:
                        last_chain = chain
                    molecule = molecules[last_chain]
                    sites = chains_res[last_chain]
                    last_molecule = molecule

                if nline == maxnlines:
                    cur_atoms.append(aname)

            if line == "TER\n":
                resnumb += 1

            if (
                prev_resnumb != resnumb or nline == maxnlines or chain != last_chain
            ) and prev_resnumb is not None:

                if nline == maxnlines:
                    prev_resnumb = copy(resnumb)
                    resnumb = "None"

                if last_chain in molecules:
                    if (
                        prev_resname in TITRABLERESIDUES
                        or (prev_resnumb in molecule.NTR or resnumb in molecule.NTR)
                        or (prev_resnumb in molecule.CTR or resnumb in molecule.CTR)
                    ):
                        if prev_resnumb in molecule.NTR and resnumb not in molecule.NTR:
                            check_site(
                                molecule,
                                prev_resnumb,
                                sites,
                                prev_resname,
                                cur_atoms,
                                ter="NTR",
                            )
                            prev_resnumb = None
                        # Dealing with the last residue and CTR
                        elif (
                            prev_resnumb in molecule.CTR and resnumb not in molecule.CTR
                        ) or (prev_resnumb in molecule.CTR and chain != last_chain):
                            check_site(
                                molecule,
                                prev_resnumb,
                                sites,
                                prev_resname,
                                cur_atoms,
                                ter="CTR",
                            )
                            prev_resnumb = None
                            last_chain = None
                        # Dealing with the previous residue
                        elif (
                            prev_resnumb is not None
                            and prev_resname in TITRABLERESIDUES
                        ):
                            check_site(
                                molecule,
                                prev_resnumb,
                                sites,
                                prev_resname,
                                cur_atoms,
                            )

                    elif prev_resname == "ALA":
                        # TODO: check residue block integrity for other non titrating residues
                        pass
                elif (
                    last_molecule
                    and prev_resnumb in last_molecule.CTR
                    and resnumb not in last_molecule.CTR
                ):
                    check_site(
                        molecule,
                        prev_resnumb,
                        sites,
                        prev_resname,
                        cur_atoms,
                        ter="CTR",
                    )
                elif prev_resname in TITRABLERESIDUES and prev_resnumb is not None:
                    check_site(molecule, prev_resnumb, sites, prev_resname, cur_atoms)

            # Dealing with the new residue
            if prev_resnumb != resnumb:
                cur_atoms = [aname]
                prev_resnumb = resnumb
                prev_resname = resname
                last_chain = chain
            elif resnumb is not None:
                cur_atoms.append(aname)
                if prev_resname in ("NTR", "CTR") and prev_resname != resname:
                    prev_resname = resname

    for molecule in molecules.values():
        # Adding the reference tautomer to each site
        molecule.addReferenceTautomers()
        # Assigning a charge set to each tautomer
        molecule.addTautomersChargeSets()

    # TODO: report blocks that failed the check (in .log file with
    # numbering reference to stepwise scheme)
    # TODO: add lipid residues
    if Config.debug:
        print("exiting check_sites_integrity")


def check_integrity(resname, res_atoms, is_titrating, ter=None):
    def read_anames(filename):
        res_atoms_st = []
        with open(filename) as f:
            for line in f.readlines()[1:]:
                cols = line.strip().split()
                aname = cols[1]
                res_atoms_st.append(aname)
        return res_atoms_st

    def pop_atoms(anames1, anames2):
        trigger = False
        for aname in anames1:
            if aname not in anames2:
                trigger = True
                if Config.debug:
                    print((aname, "not in", resname))
            else:
                anames2.remove(aname)
        if trigger:
            return anames2, False
        else:
            return anames2, True

    if Config.debug:
        print("###### INTEGRITY CHECK ######")
        print(resname)
        print(res_atoms)
    integrity = None
    integrity_ter = None

    if ter:
        st = "{}/{}tau1.st".format(Config.pypka_params["sts_path"], ter)
        res_atoms_st = read_anames(st)
        res_atoms, integrity_ter = pop_atoms(res_atoms_st, res_atoms)

    if is_titrating:
        main_chain_atoms = main_chains[Config.pypka_params["ff_family"]]
        if ter == "NTR":
            main_chain = main_chain_atoms["NTR"]
        elif ter == "CTR":
            main_chain = main_chain_atoms["CTR"]
        else:
            main_chain = main_chain_atoms["main"]

        for aname in main_chain:
            if aname in res_atoms:
                res_atoms.remove(aname)
            elif not ter:
                integrity = False

        if Config.debug:
            print(("i", integrity))

        if integrity is not False:
            filename = "{0}/{1}/sts/{2}tau1.st".format(
                Config.pypka_params["ffs_dir"], Config.pypka_params["ffID"], resname
            )
            res_atoms_st = read_anames(filename)
            res_atoms, integrity = pop_atoms(res_atoms_st, res_atoms)
            if Config.debug:
                print((res_atoms_st, res_atoms))
                print(("i", integrity))

    if len(res_atoms) != 0 and integrity:
        raise Exception("Something is wrong")

    return integrity_ter, integrity


def make_delphi_inputfile(f_in, f_out, molecules, fixed_sites):
    def getMaxCoords(coords, max_coords):
        x, y, z = coords
        max_x, max_y, max_z = max_coords
        if max_x < x:
            max_x = x
        if max_y < y:
            max_y = y
        if max_z < z:
            max_z = z
        return max_x, max_y, max_z

    def correct_termini(resnumb, resname, aname, ntr_res, ctr_res):
        if resnumb in ntr_res and aname in Config.pypka_params["NTR_atoms"]:
            resname = "NTR"
            resnumb += TERMINAL_OFFSET
        elif resnumb in ctr_res and aname in Config.pypka_params["CTR_atoms"]:
            resname = "CTR"
            resnumb += TERMINAL_OFFSET
            # if aname == "C":
            #    aname = "CT"
        return resnumb, resname, aname

    def correct_res_names(molecule, chain, resnumb, resname, aname):
        if resnumb in list(molecule.correct_names.keys()):
            resname = molecule.correct_names[resnumb]
        if (
            resnumb in list(molecule.correct_atoms.keys())
            and aname in molecule.correct_atoms[resnumb]
        ):
            aname = molecule.correct_atoms[resnumb][aname]

        if (
            Config.pypka_params["ff_family"] == "CHARMM"
            and resname == "HIS"
            # and ((fixed_sites and chain in fixed_sites and resnumb not in fixed_sites[chain]) or (fixed_sites and chain not in fixed_sites))
        ):
            resname = "HSP"

        return resnumb, resname, aname

    def assign_atoms(sites, resnumb, aname, site_Hs, site_positions):
        ref_tau_name = resname
        # print(aname in list(sites[resnumb].getRefTautomer().charge_set.keys())
        if resnumb in list(sites.keys()) and aname in list(
            sites[resnumb].getRefTautomer().charge_set.keys()
        ):
            # ( aname not in ('N', 'H', 'C', 'O', 'CA') or
            # (aname in ('N', 'H', 'C', 'O', 'CA') and resname == 'NTR')):
            # change res name to reference tautomer
            ref_tau_name = sites[resnumb].getRefTautomerName()

            # add atom to corresponding site
            sites[resnumb].addAtom(aname, anumb)

            if chain not in site_positions:
                site_positions[chain] = {}
                site_Hs[chain] = {}
            if resnumb not in site_positions[chain]:
                site_positions[chain][resnumb] = []
                site_Hs[chain][resnumb] = []

            if resnumb in site_positions[chain]:
                site_positions[chain][resnumb].append((x, y, z))
                if aname[0] == "H":
                    site_Hs[chain][resnumb].append((x, y, z))

        return site_Hs, ref_tau_name, site_positions

    new_pdb_content = ""
    site_positions = {}
    site_Hs = {}
    max_box = [0.0, 0.0, 0.0]
    aposition = -1
    sequence = {}
    with open(f_in) as f:
        for line in f:
            if line.startswith("ATOM"):
                aposition += 1
                (aname, anumb, resname, chain, resnumb, x, y, z) = read_pdb_line(line)

                max_box = getMaxCoords([x, y, z], max_box)

                if chain not in sequence:
                    sequence[chain] = {}
                if resnumb not in sequence[chain]:
                    sequence[chain][resnumb] = resname

                if chain in molecules:
                    molecule = molecules[chain]
                    ntr_res = molecule.NTR
                    ctr_res = molecule.CTR
                    sites = molecule.sites

                    if (
                        resname == "HIS"
                        and aname == "HD1"
                        and resnumb not in sites.keys()
                    ):
                        aposition -= 1
                        continue

                    resnumb, resname, aname = correct_termini(
                        resnumb, resname, aname, ntr_res, ctr_res
                    )

                    resnumb, resname, aname = correct_res_names(
                        molecule, chain, resnumb, resname, aname
                    )

                    titrable_res = False
                    if resnumb in sites.keys():
                        titrable_res = True

                    molecule.addAtom(aname, anumb, aposition, titrable_res)

                    (site_Hs, resname, site_positions) = assign_atoms(
                        sites, resnumb, aname, site_Hs, site_positions
                    )

                else:
                    if resname == "HIS" and aname == "HD1":
                        aposition -= 1
                        continue
                    resnumb, resname, aname = correct_res_names(
                        molecule, chain, resnumb, resname, aname
                    )

                new_pdb_content += new_pdb_line(
                    aposition, aname, resname, resnumb, x, y, z, chain=chain
                )

            elif line.startswith("CRYST1"):
                parts = line.split()
                box = [float(i) for i in parts[1:4]]

    if box == [1.0, 1.0, 1.0]:
        box = max_box

    if Config.pypka_params["box"]:
        box = Config.pypka_params["box"]
    else:
        Config.pypka_params.setBox(box)

    if Config.delphi_params["pbc_dim"] == 2:
        Config.delphi_params.redefineScale()

    new_pdb_content += "TER\nENDMDL\n"
    with open(f_out, "w") as f_new:
        f_new.write(new_pdb_content)

    # TODO: check terminal_offset has to be bigger than the total number of residues
    # TODO: delete terminal_offset and use another approach to distinguish between N- and C-ter
    # TODO: check size xy > config.cutoff * 2
    # if so, raise Exception, and ask to change cutoff value

    # TODO: check if pbc_dim -> set gsizes from pdb size xy and ignore perfil

    for chain in site_positions.keys():
        molecule = molecules[chain]
        for site in site_positions[chain]:
            if site in list(molecule.sites.keys()):
                pos_max = [-9999990, -999999, -999999]
                pos_min = [999999, 999999, 999999]
                focus_center = [0, 0, 0]
                for atom in site_positions[chain][site]:
                    for i in range(3):
                        if pos_max[i] < atom[i]:
                            pos_max[i] = atom[i]
                        if pos_min[i] > atom[i]:
                            pos_min[i] = atom[i]
                focus_center[0] = (pos_max[0] + pos_min[0]) / 2
                focus_center[1] = (pos_max[1] + pos_min[1]) / 2
                focus_center[2] = (pos_max[2] + pos_min[2]) / 2

                if Config.delphi_params["pbc_dim"] == 2:
                    molecule.sites[site].addCenter(
                        focus_center, boxsize=box[0], box_z=box[2]
                    )
                else:
                    molecule.sites[site].addCenter(focus_center)
                hx, hy, hz = 0, 0, 0
                nHs = len(site_Hs[chain][site])
                if nHs == 0:
                    sitename = molecule.sites[site].getName()
                    raise Exception(
                        "Site {1}{0} appears "
                        "to have no Hydrogen atoms".format(site, sitename)
                    )
                for h in site_Hs[chain][site]:
                    hx += h[0]
                    hy += h[1]
                    hz += h[2]
                hx /= nHs
                hy /= nHs
                hz /= nHs
                Hcenter = (hx, hy, hz)
                molecule.sites[site].addCenterH(Hcenter)

    return sequence


def fix_fixed_sites(molecules, fixed_sites, fname):
    for chain in fixed_sites:
        for site, state in list(fixed_sites[chain].items()):
            if isinstance(site, str) and site[-1] in "NC":
                del fixed_sites[chain][site]
                site = int(site[:-1]) + TERMINAL_OFFSET
                fixed_sites[chain][site] = state

    for molecule in molecules.values():
        chain = molecule.chain
        for sitenumb, site in list(molecule.sites.items()):
            if sitenumb in fixed_sites[chain]:
                del molecule.sites[sitenumb]
                site_i = molecule.sites_order.index(site)
                del molecule.sites_order[site_i]

    new_pdb_content = ""
    with open(fname) as f:
        for line in f:
            if line.startswith("ATOM "):
                (aname, anumb, resname, chain, resnumb, x, y, z) = read_pdb_line(line)
                if resnumb in fixed_sites[chain]:
                    resname = "{}{}".format(
                        resname[:-1], str(fixed_sites[chain][resnumb])
                    )
                new_pdb_content += new_pdb_line(
                    anumb, aname, resname, resnumb, x, y, z, chain=chain
                )
            else:
                new_pdb_content += line
    with open(fname, "w") as f:
        f.write(new_pdb_content)



rt = 15  # 2 5 'zero' 'min'

water_atype = "OW"

ignore_list = ()

rules = {
    "NTR": {
        "N": "nl",
        "H1": "h",
        "H2": "h",
        "H3": "h",
        "CA": "ch1",
        "C": "c",
        "O": "o",
    },
    "CTR": {
        "CT": "c",
        "O1": "om",
        "O2": "om",
        "HO11": "h",
        "HO12": "h",
        "HO21": "h",
        "HO22": "h",
    },
}

rules["NT0"] = rules["NTR"]
rules["NT1"] = rules["NTR"]
rules["NT2"] = rules["NTR"]
rules["NT3"] = rules["NTR"]
rules["CT0"] = rules["CTR"]
rules["CT1"] = rules["CTR"]
rules["CT2"] = rules["CTR"]
rules["CT3"] = rules["CTR"]
rules["CT4"] = rules["CTR"]


def radius_from_lennard(c6, c12, rt="min", water=False):
    def calc_radius(rmin, c6, c12, rt):
        if isinstance(rt, int):
            kT = 2.49432
            radius = (rmin ** (-6) + (rt * kT / c12) ** 0.5) ** (-1 / 6)
        elif rt == "zero":
            radius = rmin * 2 ** (-1 / 6)
        else:
            radius = rmin
        return radius

    if not c6:
        return 0.001

    rmin = (2 * c12 / c6) ** (1.0 / 6)

    radius = calc_radius(rmin, c6, c12, rt)

    if water:
        return 10 * radius / 2
    else:
        return 10 * radius - water_radius


db_lj = {}
db_lj_atomic = {}
with open("ffG54a7pHtnb.itp") as f:
    trigger = False
    for line in f:
        cols = line.strip().split()
        if not line.startswith(";") and "nonbond_params" in line:
            trigger = True
        elif "[ " in line:
            trigger = False

        if trigger and len(cols) == 5:
            if cols[0] == water_atype:
                atype = cols[1].lower()
            elif cols[1] == water_atype:
                atype = cols[0].lower()
            else:
                continue

            c6, c12 = float(cols[-2]), float(cols[-1])
            db_lj[atype] = (c6, c12)


db_res = {}
trigger = False
with open("ffG54a7pHt.rtp") as f:
    for line in f:
        if line.startswith("["):  # begin residue block
            resname = line.split(";")[0].strip("[ ]\n")
            db_res[resname] = {}
        elif "[ atoms ]" in line:
            trigger = True
        elif " [ " in line:
            trigger = False
        elif trigger and line.strip():
            cols = line.strip().split()
            aname, atype = cols[0], cols[1].lower()
            db_res[resname][aname] = atype

for i in range(4):
    rules[f"SE{i}"] = db_res["SER"]
    for ii in range(4):
        rules[f"SE{i}"][f"HG{ii}"] = db_res["SER"]["HG"]
for i in range(4):
    rules[f"TH{i}"] = db_res["THR"]
    for ii in range(4):
        rules[f"TH{i}"][f"HG{ii}"] = db_res["THR"]["HG1"]
rules["HIS"] = db_res["HI2"]
rules["LYS"] = db_res["LY3"]

db_res["LYS"] = db_res["LY3"]
db_res["LY0"] = db_res["LY3"]
db_res["LY1"] = db_res["LY3"]
db_res["LY2"] = db_res["LY3"]

db_res["GLU"] = db_res["GL4"]
db_res["GL0"] = db_res["GL4"]
db_res["GL1"] = db_res["GL4"]
db_res["GL2"] = db_res["GL4"]
db_res["GL3"] = db_res["GL4"]

db_res["ASP"] = db_res["AS4"]
db_res["AS0"] = db_res["AS4"]
db_res["AS1"] = db_res["AS4"]
db_res["AS2"] = db_res["AS4"]
db_res["AS3"] = db_res["AS4"]

db_res["ARG"] = db_res["AR4"]
db_res["AR0"] = db_res["AR4"]
db_res["AR1"] = db_res["AR4"]
db_res["AR2"] = db_res["AR4"]
db_res["AR3"] = db_res["AR4"]

# c6, c12 = db_lj[water_atype.lower()]
# water_radius = radius_from_lennard(c6, c12, rt=rt, water=True)
water_radius = 1.4241233005024403
print(
    """! Hydrogen atoms have 0.001 radii because
! DelPhi complains if H has charge and no
! radii. It does not change final results
atom__res_radius_"""
)
with open("DataBaseT_old.siz") as f:
    for line in f.readlines()[4:]:
        cols = line.strip().split()
        aname, resname = cols[0:2]

        if resname in ignore_list:
            continue

        if resname in db_res and aname in db_res[resname]:
            atype = db_res[resname][aname]
        elif resname in rules and aname in rules[resname]:
            atype = rules[resname][aname]
        else:
            print(aname, resname, "ignored")
            exit()
            continue

        c6, c12 = db_lj[atype]
        radius = radius_from_lennard(c6, c12, rt=rt)

        out = f"{aname:4}  {resname:3}   {radius:1.3f}"

        print(out)
